@@ -1,3 +1,7 @@
+// SPDX-License-Identifier: MIT OR LGPL-2.0-or-later
+// SPDX-FileCopyrightText: 2012 Red Hat, Inc.
+// SPDX-FileCopyrightText: 2013 Giovanni Campagna <gcampagna@src.gnome.org>
+
 // We use Gio to have some objects that we know exist
 const Gio = imports.gi.Gio;
 const GObject = imports.gi.GObject;
@@ -5,9 +9,9 @@ const GObject = imports.gi.GObject;
 describe('Looking up param specs', function () {
     let p1, p2;
     beforeEach(function () {
-        let find_property = GObject.Object.find_property;
-        p1 = find_property.call(Gio.ThemedIcon, 'name');
-        p2 = find_property.call(Gio.SimpleAction, 'enabled');
+        let findProperty = GObject.Object.find_property;
+        p1 = findProperty.call(Gio.ThemedIcon, 'name');
+        p2 = findProperty.call(Gio.SimpleAction, 'enabled');
     });
 
     it('works', function () {
@@ -59,13 +63,3 @@ describe('GType marshalling', function () {
         expect(GObject.type_parent(GObject.TYPE_STRING)).toBeNull();
     });
 });
-
-describe('GType prototype object', function () {
-    it('has no name', function () {
-        expect(GIRepositoryGType.name).toBeNull();
-    });
-
-    it('has a string representation', function () {
-        expect(GIRepositoryGType.toString()).toEqual('[object GType prototype]');
-    });
-});
