@@ -5,7 +5,7 @@
 
 const Gi = imports._gi;
 const {CjsPrivate, GLib} = imports.gi;
-const {_checkAccessors} = imports._common;
+const {_checkAccessors, _registerType} = imports._common;
 const Legacy = imports._legacy;
 
 let GObject;
@@ -70,12 +70,56 @@ function registerClass(...args) {
             `class (is ${Object.getPrototypeOf(klass).name})`);
     }
 
+    if ('_classInit' in klass) {
+        klass = klass._classInit(klass);
+    } else {
+        // Lang.Class compatibility.
+        klass = _resolveLegacyClassFunction(klass, '_classInit')(klass);
+    }
+
+    return klass;
+}
+
+function _resolveLegacyClassFunction(klass, func) {
     // Find the "least derived" class with a _classInit static function; there
     // definitely is one, since this class must inherit from GObject
     let initclass = klass;
-    while (typeof initclass._classInit === 'undefined')
+    while (typeof initclass[func] === 'undefined')
         initclass = Object.getPrototypeOf(initclass.prototype).constructor;
-    return initclass._classInit(klass);
+    return initclass[func];
+}
+
+function _defineGType(klass, giPrototype, registeredType) {
+    const config = {
+        enumerable: false,
+        configurable: false,
+    };
+
+    /**
+     * class Example {
+     *     // The JS object for this class' ObjectPrototype
+     *     static [Gi.gobject_prototype_symbol] = ...
+     *     static get $gtype () {
+     *         return ...;
+     *     }
+     * }
+     *
+     * // Equal to the same property on the constructor
+     * Example.prototype[Gi.gobject_prototype_symbol] = ...
+     */
+
+    Object.defineProperty(klass, '$gtype', {
+        ...config,
+        get() {
+            return registeredType;
+        },
+    });
+
+    Object.defineProperty(klass.prototype, Gi.gobject_prototype_symbol, {
+        ...config,
+        writable: false,
+        value: giPrototype,
+    });
 }
 
 // Some common functions between GObject.Class and GObject.Interface
@@ -166,14 +210,21 @@ function _propertiesAsArray(klass) {
     return propertiesArray;
 }
 
-function _copyAllDescriptors(target, source, filter) {
-    Object.getOwnPropertyNames(source)
-    .filter(key => !['prototype', 'constructor'].concat(filter).includes(key))
-    .concat(Object.getOwnPropertySymbols(source))
-    .forEach(key => {
-        let descriptor = Object.getOwnPropertyDescriptor(source, key);
-        Object.defineProperty(target, key, descriptor);
-    });
+function _copyInterfacePrototypeDescriptors(targetPrototype, sourceInterface) {
+    Object.entries(Object.getOwnPropertyDescriptors(sourceInterface))
+        .filter(([key, descriptor]) =>
+            // Don't attempt to copy the constructor or toString implementations
+            !['constructor', 'toString'].includes(key) &&
+            // Ignore properties starting with __
+            (typeof key !== 'string' || !key.startsWith('__')) &&
+            // Don't override an implementation on the target
+            !targetPrototype.hasOwnProperty(key) &&
+            descriptor &&
+            // Only copy if the descriptor has a getter, is a function, or is enumerable.
+            (typeof descriptor.value === 'function' || descriptor.get || descriptor.enumerable))
+        .forEach(([key, descriptor]) => {
+            Object.defineProperty(targetPrototype, key, descriptor);
+        });
 }
 
 function _interfacePresent(required, klass) {
@@ -222,6 +273,14 @@ function _checkInterface(iface, proto) {
     }
 }
 
+function _checkProperties(klass) {
+    if (!klass.hasOwnProperty(properties))
+        return;
+
+    for (let pspec of Object.values(klass[properties]))
+        _checkAccessors(klass.prototype, pspec, GObject);
+}
+
 function _init() {
     GObject = this;
 
@@ -431,7 +490,45 @@ function _init() {
 
     GObject.registerClass = registerClass;
 
+    GObject.Object.new = function (gtype, props = {}) {
+        const constructor = Gi.lookupConstructor(gtype);
+
+        if (!constructor)
+            throw new Error(`Constructor for gtype ${gtype} not found`);
+        return new constructor(props);
+    };
+
+    GObject.Object.new_with_properties = function (gtype, names, values) {
+        if (!Array.isArray(names) || !Array.isArray(values))
+            throw new Error('new_with_properties takes two arrays (names, values)');
+        if (names.length !== values.length)
+            throw new Error('Arrays passed to new_with_properties must be the same length');
+
+        const props = Object.fromEntries(names.map((name, ix) => [name, values[ix]]));
+        return GObject.Object.new(gtype, props);
+    };
+
     GObject.Object._classInit = function (klass) {
+        _checkProperties(klass);
+
+        if (_registerType in klass)
+            klass[_registerType]();
+        else
+            _resolveLegacyClassFunction(klass, _registerType).call(klass);
+
+        return klass;
+    };
+
+    // For backwards compatibility only. Use instanceof instead.
+    GObject.Object.implements = function (iface) {
+        if (iface.$gtype)
+            return GObject.type_is_a(this, iface.$gtype);
+        return false;
+    };
+
+    function registerGObjectType() {
+        let klass = this;
+
         let gtypename = _createGTypeName(klass);
         let gflags = klass.hasOwnProperty(GTypeFlags) ? klass[GTypeFlags] : 0;
         let gobjectInterfaces = klass.hasOwnProperty(interfaces) ? klass[interfaces] : [];
@@ -439,36 +536,39 @@ function _init() {
         let parent = Object.getPrototypeOf(klass);
         let gobjectSignals = klass.hasOwnProperty(signals) ? klass[signals] : [];
 
-        propertiesArray.forEach(pspec => _checkAccessors(klass.prototype, pspec, GObject));
+        // Default to the GObject-specific prototype, fallback on the JS prototype for GI native classes.
+        const parentPrototype = parent.prototype[Gi.gobject_prototype_symbol] ?? parent.prototype;
 
-        let newClass = Gi.register_type(parent.prototype, gtypename, gflags,
+        const [giPrototype, registeredType] = Gi.register_type_with_class(
+            klass, parentPrototype, gtypename, gflags,
             gobjectInterfaces, propertiesArray);
-        Object.setPrototypeOf(newClass, parent);
 
-        _createSignals(newClass.$gtype, gobjectSignals);
+        _defineGType(klass, giPrototype, registeredType);
+        _createSignals(klass.$gtype, gobjectSignals);
 
-        _copyAllDescriptors(newClass, klass);
-        gobjectInterfaces.forEach(iface =>
-            _copyAllDescriptors(newClass.prototype, iface.prototype,
-                ['toString']));
-        _copyAllDescriptors(newClass.prototype, klass.prototype);
+        // Reverse the interface array to give the last required interface precedence over the first.
+        const requiredInterfaces = [...gobjectInterfaces].reverse();
+        requiredInterfaces.forEach(iface =>
+            _copyInterfacePrototypeDescriptors(klass, iface));
+        requiredInterfaces.forEach(iface =>
+            _copyInterfacePrototypeDescriptors(klass.prototype, iface.prototype));
 
-        Object.getOwnPropertyNames(newClass.prototype)
+        Object.getOwnPropertyNames(klass.prototype)
         .filter(name => name.startsWith('vfunc_') || name.startsWith('on_'))
         .forEach(name => {
-            let descr = Object.getOwnPropertyDescriptor(newClass.prototype, name);
+            let descr = Object.getOwnPropertyDescriptor(klass.prototype, name);
             if (typeof descr.value !== 'function')
                 return;
 
-            let func = newClass.prototype[name];
+            let func = klass.prototype[name];
 
             if (name.startsWith('vfunc_')) {
-                newClass.prototype[Gi.hook_up_vfunc_symbol](name.slice(6), func);
+                giPrototype[Gi.hook_up_vfunc_symbol](name.slice(6), func);
             } else if (name.startsWith('on_')) {
                 let id = GObject.signal_lookup(name.slice(3).replace('_', '-'),
-                    newClass.$gtype);
+                    klass.$gtype);
                 if (id !== 0) {
-                    GObject.signal_override_class_closure(id, newClass.$gtype, function (...argArray) {
+                    GObject.signal_override_class_closure(id, klass.$gtype, function (...argArray) {
                         let emitter = argArray.shift();
 
                         return func.apply(emitter, argArray);
@@ -478,30 +578,61 @@ function _init() {
         });
 
         gobjectInterfaces.forEach(iface =>
-            _checkInterface(iface, newClass.prototype));
+            _checkInterface(iface, klass.prototype));
 
-        // For backwards compatibility only. Use instanceof instead.
-        newClass.implements = function (iface) {
-            if (iface.$gtype)
-                return GObject.type_is_a(newClass.$gtype, iface.$gtype);
-            return false;
-        };
+        // Lang.Class parent classes don't support static inheritance
+        if (!('implements' in klass))
+            klass.implements = GObject.Object.implements;
+    }
 
-        return newClass;
-    };
+    Object.defineProperty(GObject.Object, _registerType, {
+        value: registerGObjectType,
+        writable: false,
+        configurable: false,
+        enumerable: false,
+    });
+
+    function interfaceInstanceOf(instance) {
+        if (instance && typeof instance === 'object' &&
+            GObject.Interface.prototype.isPrototypeOf(this.prototype))
+            return GObject.type_is_a(instance, this);
+
+        return false;
+    }
+
+    function registerInterfaceType() {
+        let klass = this;
 
-    GObject.Interface._classInit = function (klass) {
         let gtypename = _createGTypeName(klass);
         let gobjectInterfaces = klass.hasOwnProperty(requires) ? klass[requires] : [];
         let props = _propertiesAsArray(klass);
         let gobjectSignals = klass.hasOwnProperty(signals) ? klass[signals] : [];
 
-        let newInterface = Gi.register_interface(gtypename, gobjectInterfaces,
+        const [giPrototype, registeredType] = Gi.register_interface_with_class(klass, gtypename, gobjectInterfaces,
             props);
 
-        _createSignals(newInterface.$gtype, gobjectSignals);
+        _defineGType(klass, giPrototype, registeredType);
+        _createSignals(klass.$gtype, gobjectSignals);
 
-        _copyAllDescriptors(newInterface, klass);
+        Object.defineProperty(klass, Symbol.hasInstance, {
+            value: interfaceInstanceOf,
+        });
+
+        return klass;
+    }
+
+    Object.defineProperty(GObject.Interface, _registerType, {
+        value: registerInterfaceType,
+        writable: false,
+        configurable: false,
+        enumerable: false,
+    });
+
+    GObject.Interface._classInit = function (klass) {
+        if (_registerType in klass)
+            klass[_registerType]();
+        else
+            _resolveLegacyClassFunction(klass, _registerType).call(klass);
 
         Object.getOwnPropertyNames(klass.prototype)
         .filter(key => key !== 'constructor')
@@ -514,15 +645,15 @@ function _init() {
             // SomeInterface.prototype.some_function.call(this, blah)
             if (typeof descr.value === 'function') {
                 let interfaceProto = klass.prototype;  // capture in closure
-                newInterface[key] = function (thisObj, ...args) {
+                klass[key] = function (thisObj, ...args) {
                     return interfaceProto[key].call(thisObj, ...args);
                 };
             }
 
-            Object.defineProperty(newInterface.prototype, key, descr);
+            Object.defineProperty(klass.prototype, key, descr);
         });
 
-        return newInterface;
+        return klass;
     };
 
     /**
@@ -560,6 +691,12 @@ function _init() {
         return CjsPrivate.g_object_bind_property_full(this, ...args);
     };
 
+    if (GObject.BindingGroup !== undefined) {
+        GObject.BindingGroup.prototype.bind_full = function (...args) {
+            return CjsPrivate.g_binding_group_bind_full(this, ...args);
+        };
+    }
+
     // fake enum for signal accumulators, keep in sync with gi/object.c
     GObject.AccumulatorType = {
         NONE: 0,
