@@ -19,10 +19,14 @@
 
 #include <js/BigInt.h>
 #include <js/CharacterEncoding.h>
+#include <js/CompilationAndEvaluation.h>
+#include <js/CompileOptions.h>
 #include <js/Exception.h>
 #include <js/Id.h>
 #include <js/PropertyAndElement.h>
+#include <js/Realm.h>
 #include <js/RootingAPI.h>
+#include <js/SourceText.h>
 #include <js/String.h>
 #include <js/TypeDecls.h>
 #include <js/Utility.h>  // for UniqueChars
@@ -41,6 +45,10 @@
 #include "test/gjs-test-utils.h"
 #include "util/misc.h"
 
+namespace mozilla {
+union Utf8Unit;
+}
+
 // COMPAT: https://gitlab.gnome.org/GNOME/glib/-/merge_requests/1553
 #ifdef __clang_analyzer__
 void g_assertion_message(const char*, const char*, int, const char*,
@@ -117,7 +125,7 @@ gjstest_test_func_gjs_context_construct_eval(void)
 {
     GjsContext *context;
     int estatus;
-    GError *error = NULL;
+    GjsAutoError error;
 
     context = gjs_context_new ();
     if (!gjs_context_eval (context, "1+1", -1, "<input>", &estatus, &error))
@@ -127,7 +135,7 @@ gjstest_test_func_gjs_context_construct_eval(void)
 
 static void gjstest_test_func_gjs_context_eval_dynamic_import() {
     GjsAutoUnref<GjsContext> gjs = gjs_context_new();
-    GError* error = NULL;
+    GjsAutoError error;
     int status;
 
     bool ok = gjs_context_eval(gjs, R"js(
@@ -144,7 +152,7 @@ static void gjstest_test_func_gjs_context_eval_dynamic_import() {
 
 static void gjstest_test_func_gjs_context_eval_dynamic_import_relative() {
     GjsAutoUnref<GjsContext> gjs = gjs_context_new();
-    GError* error = NULL;
+    GjsAutoError error;
     int status;
 
     bool ok = g_file_set_contents("num.js", "export default 77;", -1, &error);
@@ -172,7 +180,7 @@ static void gjstest_test_func_gjs_context_eval_dynamic_import_relative() {
 
 static void gjstest_test_func_gjs_context_eval_dynamic_import_bad() {
     GjsAutoUnref<GjsContext> gjs = gjs_context_new();
-    GError* error = NULL;
+    GjsAutoError error;
     int status;
 
     g_test_expect_message("Gjs", G_LOG_LEVEL_WARNING,
@@ -196,13 +204,11 @@ static void gjstest_test_func_gjs_context_eval_dynamic_import_bad() {
     g_assert_cmpuint(status, ==, 10);
 
     g_test_assert_expected_messages();
-
-    g_clear_error(&error);
 }
 
 static void gjstest_test_func_gjs_context_eval_non_zero_terminated(void) {
     GjsAutoUnref<GjsContext> gjs = gjs_context_new();
-    GError* error = NULL;
+    GjsAutoError error;
     int status;
 
     // This string is invalid JS if it is treated as zero-terminated
@@ -217,7 +223,7 @@ static void
 gjstest_test_func_gjs_context_exit(void)
 {
     GjsContext *context = gjs_context_new();
-    GError *error = NULL;
+    GjsAutoError error;
     int status;
 
     bool ok = gjs_context_eval(context, "imports.system.exit(0);", -1,
@@ -226,7 +232,7 @@ gjstest_test_func_gjs_context_exit(void)
     g_assert_error(error, GJS_ERROR, GJS_ERROR_SYSTEM_EXIT);
     g_assert_cmpuint(status, ==, 0);
 
-    g_clear_error(&error);
+    error.reset();
 
     ok = gjs_context_eval(context, "imports.system.exit(42);", -1, "<input>",
                           &status, &error);
@@ -234,14 +240,13 @@ gjstest_test_func_gjs_context_exit(void)
     g_assert_error(error, GJS_ERROR, GJS_ERROR_SYSTEM_EXIT);
     g_assert_cmpuint(status, ==, 42);
 
-    g_clear_error(&error);
     g_object_unref(context);
 }
 
 static void gjstest_test_func_gjs_context_eval_module_file() {
     GjsAutoUnref<GjsContext> gjs = gjs_context_new();
     uint8_t exit_status;
-    GError* error = nullptr;
+    GjsAutoError error;
 
     bool ok = gjs_context_eval_module_file(
         gjs, "resource:///org/gnome/gjs/mock/test/modules/default.js",
@@ -256,7 +261,7 @@ static void gjstest_test_func_gjs_context_eval_module_file() {
 static void gjstest_test_func_gjs_context_eval_module_file_throw() {
     GjsAutoUnref<GjsContext> gjs = gjs_context_new();
     uint8_t exit_status;
-    GError* error = nullptr;
+    GjsAutoError error;
 
     g_test_expect_message("Gjs", G_LOG_LEVEL_CRITICAL, "*bad module*");
 
@@ -269,13 +274,11 @@ static void gjstest_test_func_gjs_context_eval_module_file_throw() {
     g_assert_cmpuint(exit_status, ==, 1);
 
     g_test_assert_expected_messages();
-
-    g_clear_error(&error);
 }
 
 static void gjstest_test_func_gjs_context_eval_module_file_exit() {
     GjsAutoUnref<GjsContext> gjs = gjs_context_new();
-    GError* error = nullptr;
+    GjsAutoError error;
     uint8_t exit_status;
 
     bool ok = gjs_context_eval_module_file(
@@ -286,7 +289,7 @@ static void gjstest_test_func_gjs_context_eval_module_file_exit() {
     g_assert_error(error, GJS_ERROR, GJS_ERROR_SYSTEM_EXIT);
     g_assert_cmpuint(exit_status, ==, 0);
 
-    g_clear_error(&error);
+    error.reset();
 
     ok = gjs_context_eval_module_file(
         gjs, "resource:///org/gnome/gjs/mock/test/modules/exit.js",
@@ -295,19 +298,17 @@ static void gjstest_test_func_gjs_context_eval_module_file_exit() {
     g_assert_false(ok);
     g_assert_error(error, GJS_ERROR, GJS_ERROR_SYSTEM_EXIT);
     g_assert_cmpuint(exit_status, ==, 42);
-
-    g_clear_error(&error);
 }
 
 static void gjstest_test_func_gjs_context_eval_module_file_fail_instantiate() {
     GjsAutoUnref<GjsContext> gjs = gjs_context_new();
-    GError* error = nullptr;
+    GjsAutoError error;
     uint8_t exit_status;
 
     g_test_expect_message("Gjs", G_LOG_LEVEL_WARNING, "*foo*");
 
     // evaluating this module without registering 'foo' first should make it
-    // fail ModuleInstantiate
+    // fail ModuleLink
     bool ok = gjs_context_eval_module_file(
         gjs, "resource:///org/gnome/gjs/mock/test/modules/import.js",
         &exit_status, &error);
@@ -317,13 +318,11 @@ static void gjstest_test_func_gjs_context_eval_module_file_fail_instantiate() {
     g_assert_cmpuint(exit_status, ==, 1);
 
     g_test_assert_expected_messages();
-
-    g_clear_error(&error);
 }
 
 static void gjstest_test_func_gjs_context_eval_module_file_exit_code_omitted_warning() {
     GjsAutoUnref<GjsContext> gjs = gjs_context_new();
-    GError* error = nullptr;
+    GjsAutoError error;
 
     g_test_expect_message("Gjs", G_LOG_LEVEL_WARNING, "*foo*");
 
@@ -335,14 +334,12 @@ static void gjstest_test_func_gjs_context_eval_module_file_exit_code_omitted_war
     g_assert_error(error, GJS_ERROR, GJS_ERROR_FAILED);
 
     g_test_assert_expected_messages();
-
-    g_clear_error(&error);
 }
 
 static void
 gjstest_test_func_gjs_context_eval_module_file_exit_code_omitted_no_warning() {
     GjsAutoUnref<GjsContext> gjs = gjs_context_new();
-    GError* error = nullptr;
+    GjsAutoError error;
 
     bool ok = gjs_context_eval_module_file(
         gjs, "resource:///org/gnome/gjs/mock/test/modules/default.js", nullptr,
@@ -350,12 +347,11 @@ gjstest_test_func_gjs_context_eval_module_file_exit_code_omitted_no_warning() {
 
     g_assert_true(ok);
     g_assert_no_error(error);
-    g_clear_error(&error);
 }
 
 static void gjstest_test_func_gjs_context_eval_file_exit_code_omitted_throw() {
     GjsAutoUnref<GjsContext> gjs = gjs_context_new();
-    GError* error = nullptr;
+    GjsAutoError error;
 
     g_test_expect_message("Gjs", G_LOG_LEVEL_CRITICAL, "*bad module*");
 
@@ -367,13 +363,11 @@ static void gjstest_test_func_gjs_context_eval_file_exit_code_omitted_throw() {
     g_assert_error(error, GJS_ERROR, GJS_ERROR_FAILED);
 
     g_test_assert_expected_messages();
-
-    g_clear_error(&error);
 }
 
 static void gjstest_test_func_gjs_context_eval_file_exit_code_omitted_no_throw() {
     GjsAutoUnref<GjsContext> gjs = gjs_context_new();
-    GError* error = nullptr;
+    GjsAutoError error;
 
     bool ok = gjs_context_eval_file(
         gjs, "resource:///org/gnome/gjs/mock/test/modules/nothrows.js", nullptr,
@@ -381,13 +375,11 @@ static void gjstest_test_func_gjs_context_eval_file_exit_code_omitted_no_throw()
 
     g_assert_true(ok);
     g_assert_no_error(error);
-
-    g_clear_error(&error);
 }
 
 static void gjstest_test_func_gjs_context_register_module_eval_module() {
     GjsAutoUnref<GjsContext> gjs = gjs_context_new();
-    GError* error = nullptr;
+    GjsAutoError error;
 
     bool ok = gjs_context_register_module(
         gjs, "foo", "resource:///org/gnome/gjs/mock/test/modules/default.js",
@@ -406,7 +398,7 @@ static void gjstest_test_func_gjs_context_register_module_eval_module() {
 
 static void gjstest_test_func_gjs_context_register_module_eval_module_file() {
     GjsAutoUnref<GjsContext> gjs = gjs_context_new();
-    GError* error = nullptr;
+    GjsAutoError error;
 
     bool ok = gjs_context_register_module(
         gjs, "foo", "resource:///org/gnome/gjs/mock/test/modules/default.js",
@@ -425,21 +417,83 @@ static void gjstest_test_func_gjs_context_register_module_eval_module_file() {
     g_assert_cmpuint(exit_status, ==, 0);
 }
 
+static void gjstest_test_func_gjs_context_register_module_eval_jsapi(
+    GjsUnitTestFixture* fx, const void*) {
+    GjsAutoError error;
+
+    bool ok = gjs_context_register_module(
+        fx->gjs_context, "foo",
+        "resource:///org/gnome/gjs/mock/test/modules/default.js", &error);
+    g_assert_true(ok);
+    g_assert_no_error(error);
+
+    JS::CompileOptions options{fx->cx};
+    options.setFileAndLine("import.js", 1);
+    static const char* code = R"js(
+        let error;
+        const loop = new imports.gi.GLib.MainLoop(null, false);
+        import('foo')
+        .then(module => {
+            if (module.default !== 77)
+                throw new Error('wrong number');
+        })
+        .catch(e => (error = e))
+        .finally(() => loop.quit());
+        loop.run();
+        if (error)
+            throw error;
+    )js";
+    JS::SourceText<mozilla::Utf8Unit> source;
+    ok = source.init(fx->cx, code, strlen(code), JS::SourceOwnership::Borrowed);
+    g_assert_true(ok);
+
+    JS::RootedValue unused{fx->cx};
+    ok = JS::Evaluate(fx->cx, options, source, &unused);
+    gjs_log_exception(fx->cx);  // will fail test if exception pending
+    g_assert_true(ok);
+}
+
+static void gjstest_test_func_gjs_context_register_module_eval_jsapi_rel(
+    GjsUnitTestFixture* fx, const void*) {
+    JS::CompileOptions options{fx->cx};
+    options.setFileAndLine("import.js", 1);
+    static const char* code = R"js(
+        let error;
+        const loop = new imports.gi.GLib.MainLoop(null, false);
+        import('./foo.js')
+        .catch(e => (error = e))
+        .finally(() => loop.quit());
+        loop.run();
+        if (error)
+            throw error;
+    )js";
+    JS::SourceText<mozilla::Utf8Unit> source;
+    bool ok =
+        source.init(fx->cx, code, strlen(code), JS::SourceOwnership::Borrowed);
+    g_assert_true(ok);
+
+    JS::RootedValue unused{fx->cx};
+    ok = JS::Evaluate(fx->cx, options, source, &unused);
+    g_assert_false(ok);
+    g_test_expect_message("Gjs", G_LOG_LEVEL_WARNING,
+                          "JS ERROR: ImportError*relative*");
+    gjs_log_exception(fx->cx);
+    g_test_assert_expected_messages();
+}
+
 static void gjstest_test_func_gjs_context_register_module_non_existent() {
     GjsAutoUnref<GjsContext> gjs = gjs_context_new();
-    GError* error = nullptr;
+    GjsAutoError error;
 
     bool ok = gjs_context_register_module(gjs, "foo", "nonexist.js", &error);
 
     g_assert_false(ok);
     g_assert_error(error, GJS_ERROR, GJS_ERROR_FAILED);
-
-    g_clear_error(&error);
 }
 
 static void gjstest_test_func_gjs_context_eval_module_unregistered() {
     GjsAutoUnref<GjsContext> gjs = gjs_context_new();
-    GError* error = nullptr;
+    GjsAutoError error;
     uint8_t exit_status;
 
     bool ok = gjs_context_eval_module(gjs, "foo", &exit_status, &error);
@@ -447,25 +501,21 @@ static void gjstest_test_func_gjs_context_eval_module_unregistered() {
     g_assert_false(ok);
     g_assert_error(error, GJS_ERROR, GJS_ERROR_FAILED);
     g_assert_cmpuint(exit_status, ==, 1);
-
-    g_clear_error(&error);
 }
 
 static void gjstest_test_func_gjs_context_eval_module_exit_code_omitted_throw() {
     GjsAutoUnref<GjsContext> gjs = gjs_context_new();
-    GError* error = nullptr;
+    GjsAutoError error;
 
     bool ok = gjs_context_eval_module(gjs, "foo", nullptr, &error);
 
     g_assert_false(ok);
     g_assert_error(error, GJS_ERROR, GJS_ERROR_FAILED);
-
-    g_clear_error(&error);
 }
 
 static void gjstest_test_func_gjs_context_eval_module_exit_code_omitted_no_throw() {
     GjsAutoUnref<GjsContext> gjs = gjs_context_new();
-    GError* error = nullptr;
+    GjsAutoError error;
 
     bool ok = gjs_context_register_module(
         gjs, "lies", "resource:///org/gnome/gjs/mock/test/modules/nothrows.js",
@@ -478,8 +528,79 @@ static void gjstest_test_func_gjs_context_eval_module_exit_code_omitted_no_throw
 
     g_assert_true(ok);
     g_assert_no_error(error);
+}
 
-    g_clear_error(&error);
+static void gjstest_test_func_gjs_context_module_eval_jsapi_throws(
+    GjsUnitTestFixture* fx, const void*) {
+    GjsAutoError error;
+
+    bool ok = gjs_context_register_module(
+        fx->gjs_context, "foo",
+        "resource:///org/gnome/gjs/mock/test/modules/throws.js", &error);
+    g_assert_true(ok);
+    g_assert_no_error(error);
+
+    JS::CompileOptions options{fx->cx};
+    options.setFileAndLine("import.js", 1);
+    static const char* code = R"js(
+        let error;
+        const loop = new imports.gi.GLib.MainLoop(null, false);
+        import('foo')
+        .catch(e => (error = e))
+        .finally(() => loop.quit());
+        loop.run();
+        error;
+    )js";
+    JS::SourceText<mozilla::Utf8Unit> source;
+    ok = source.init(fx->cx, code, strlen(code), JS::SourceOwnership::Borrowed);
+    g_assert_true(ok);
+
+    JS::RootedValue thrown{fx->cx};
+    ok = JS::Evaluate(fx->cx, options, source, &thrown);
+    gjs_log_exception(fx->cx);  // will fail test if exception pending
+
+    g_assert_true(ok);
+
+    g_assert_true(thrown.isObject());
+    JS::RootedObject thrown_obj{fx->cx, &thrown.toObject()};
+    JS::RootedValue message{fx->cx};
+    ok = JS_GetProperty(fx->cx, thrown_obj, "message", &message);
+    g_assert_true(ok);
+    g_assert_true(message.isString());
+    bool match = false;
+    ok = JS_StringEqualsAscii(fx->cx, message.toString(), "bad module", &match);
+    g_assert_true(ok);
+    g_assert_true(match);
+}
+
+static void gjstest_test_func_gjs_context_run_in_realm() {
+    GjsAutoUnref<GjsContext> gjs = gjs_context_new();
+
+    auto* cx = static_cast<JSContext*>(gjs_context_get_native_context(gjs));
+    g_assert_null(JS::GetCurrentRealmOrNull(cx));
+
+    struct RunInRealmData {
+        int sentinel;
+        bool has_run;
+    } data{42, false};
+
+    gjs_context_run_in_realm(
+        gjs,
+        [](GjsContext* gjs, void* ptr) {
+            g_assert_true(GJS_IS_CONTEXT(gjs));
+            auto* data = static_cast<RunInRealmData*>(ptr);
+            g_assert_cmpint(data->sentinel, ==, 42);
+
+            auto* cx =
+                static_cast<JSContext*>(gjs_context_get_native_context(gjs));
+            g_assert_nonnull(JS::GetCurrentRealmOrNull(cx));
+
+            data->has_run = true;
+        },
+        &data);
+
+    g_assert_null(JS::GetCurrentRealmOrNull(cx));
+    g_assert_true(data.has_run);
 }
 
 #define JS_CLASS "\
@@ -491,7 +612,7 @@ static void
 gjstest_test_func_gjs_gobject_js_defined_type(void)
 {
     GjsContext *context = gjs_context_new();
-    GError *error = NULL;
+    GjsAutoError error;
     int status;
     bool ok = gjs_context_eval(context, JS_CLASS, -1, "<input>", &status, &error);
     g_assert_no_error(error);
@@ -509,7 +630,7 @@ gjstest_test_func_gjs_gobject_js_defined_type(void)
 
 static void gjstest_test_func_gjs_gobject_without_introspection(void) {
     GjsAutoUnref<GjsContext> context = gjs_context_new();
-    GError* error = nullptr;
+    GjsAutoError error;
     int status;
 
     /* Ensure class */
@@ -537,7 +658,7 @@ static void gjstest_test_func_gjs_gobject_without_introspection(void) {
 
 static void gjstest_test_func_gjs_context_eval_exit_code_omitted_throw() {
     GjsAutoUnref<GjsContext> context = gjs_context_new();
-    GError* error = nullptr;
+    GjsAutoError error;
 
     g_test_expect_message("Gjs", G_LOG_LEVEL_CRITICAL, "*wrong code*");
 
@@ -549,13 +670,11 @@ static void gjstest_test_func_gjs_context_eval_exit_code_omitted_throw() {
     g_assert_error(error, GJS_ERROR, GJS_ERROR_FAILED);
 
     g_test_assert_expected_messages();
-
-    g_clear_error(&error);
 }
 
 static void gjstest_test_func_gjs_context_eval_exit_code_omitted_no_throw() {
     GjsAutoUnref<GjsContext> context = gjs_context_new();
-    GError* error = nullptr;
+    GjsAutoError error;
 
     const char good_js[] = "let num = 77;";
 
@@ -564,8 +683,6 @@ static void gjstest_test_func_gjs_context_eval_exit_code_omitted_no_throw() {
 
     g_assert_true(ok);
     g_assert_no_error(error);
-
-    g_clear_error(&error);
 }
 
 static void gjstest_test_func_gjs_jsapi_util_string_js_string_utf8(
@@ -822,7 +939,7 @@ gjstest_test_profiler_start_stop(void)
     gjs_profiler_start(profiler);
 
     for (size_t ix = 0; ix < 100; ix++) {
-        GError *error = nullptr;
+        GjsAutoError error;
         int estatus;
 
 #define TESTJS "[1,5,7,1,2,3,67,8].sort()"
@@ -1012,14 +1129,13 @@ static void gjstest_test_args_rounded_values() {
 
 static void gjstest_test_func_gjs_context_argv_array() {
     GjsAutoUnref<GjsContext> gjs = gjs_context_new();
-    GError* error = NULL;
+    GjsAutoError error;
     int status;
 
     const char* argv[1] = {"test"};
     bool ok = gjs_context_define_string_array(gjs, "ARGV", 1, argv, &error);
 
     g_assert_no_error(error);
-    g_clear_error(&error);
     g_assert_true(ok);
 
     ok = gjs_context_eval(gjs, R"js(
@@ -1091,6 +1207,14 @@ main(int    argc,
     g_test_add_func(
         "/gjs/context/register-module/eval-module-file",
         gjstest_test_func_gjs_context_register_module_eval_module_file);
+    g_test_add("/gjs/context/register-module/eval-jsapi", GjsUnitTestFixture,
+               nullptr, gjs_unit_test_fixture_setup,
+               gjstest_test_func_gjs_context_register_module_eval_jsapi,
+               gjs_unit_test_fixture_teardown);
+    g_test_add("/gjs/context/register-module/eval-jsapi-relative",
+               GjsUnitTestFixture, nullptr, gjs_unit_test_fixture_setup,
+               gjstest_test_func_gjs_context_register_module_eval_jsapi_rel,
+               gjs_unit_test_fixture_teardown);
     g_test_add_func("/gjs/context/register-module/non-existent",
                     gjstest_test_func_gjs_context_register_module_non_existent);
     g_test_add_func("/gjs/context/eval-module/unregistered",
@@ -1127,6 +1251,12 @@ main(int    argc,
     g_test_add_func(
         "/gjs/context/eval-module/exit-code-omitted-no-throw",
         gjstest_test_func_gjs_context_eval_module_exit_code_omitted_no_throw);
+    g_test_add("/gjs/context/eval-module/jsapi-throw", GjsUnitTestFixture,
+               nullptr, gjs_unit_test_fixture_setup,
+               gjstest_test_func_gjs_context_module_eval_jsapi_throws,
+               gjs_unit_test_fixture_teardown);
+    g_test_add_func("/gjs/context/run-in-realm",
+                    gjstest_test_func_gjs_context_run_in_realm);
 
 #define ADD_JSAPI_UTIL_TEST(path, func)                            \
     g_test_add("/gjs/jsapi/util/" path, GjsUnitTestFixture, NULL,  \
